/*
 * Lcd.h
 *
 * Created: 26.01.2014 14:56:45
 *  Author: Danny
 */ 

#include <string.h>

/* Port expander registers */
#define MCP23017_IOCON_BANK0 0x0A
#define MCP23017_IOCON_BANK1 0x15

/* Register addresses in Bank 1 only */
#define MCP2307_GPIOA 0x09
#define MCP2307_IODIRB 0x10
#define MCP2307_GPIOB 0x19

/* Port expander input pin */
#define SELECT 0
#define RIGHT 1
#define DOWN 2
#define UP 3
#define Left 4

/* LED colors */
#define OFF 0x00
#define RED 0x01
#define GREEN 0x02
#define BLUE 0x04
#define YELLOW RED + BLUE
// Todo: other colors

/* LCD commands */
#define LCD_CLEARDISPLAY 0x01
#define LCD_RETURNHOME 0x02
#define LCD_ENTRYMODESET 0x04
#define LCD_DISPLAYCONTROL 0x08
#define LCD_CURSORSHIFT 0x10
#define LCD_FUNCTIONSET 0x20
#define LCD_SETCGRAMADDR 0x40
#define LCD_SETDDRAMADDR 0x80

/* Display on/off commands */
#define LCD_DISPLAYON 0x04
#define LCD_DISPLAYOFF 0x00
#define LCD_CURSORON 0x02
#define LCD_CURSOROFF 0x00
#define LCD_BLINKON 0x01
#define LCD_BLINKOFF 0x00

/* Display entry mode */
#define LCD_ENTRYRIGHT 0x00
#define LCD_ENTRYLEFT 0x02
#define LCD_ENTRYSHIFTINCREMENT 0x01
#define LCD_ENTRYSHIFTDECREMENT 0x00

/* Display / Cursor shift */
#define LCD_DISPLAYMOVE 0x08
#define LCD_CURSORMOVE 0x00
#define LCD_MOVERIGHT 0x04
#define LCD_MOVELEFT 0x00

uint8_t ddrb = 0b00010000;
uint8_t porta = 0;
uint8_t portb = 0;
uint8_t numlines = 0;

uint8_t flip[16] = { 
	0b00000000, 0b00010000, 0b00001000, 0b00011000, 
	0b00000100, 0b00010100, 0b00001100, 0b00011100, 
	0b00000010, 0b00010010, 0b00001010, 0b00011010, 
	0b00000110, 0b00010110, 0b00001110, 0b00011110 };
	
uint8_t displayshift = LCD_CURSORMOVE | LCD_MOVERIGHT;
uint8_t displaymode = LCD_ENTRYLEFT | LCD_ENTRYSHIFTDECREMENT;
uint8_t displaycontrol = LCD_DISPLAYON | LCD_CURSOROFF | LCD_BLINKOFF;


uint8_t* Out_Four(uint8_t bitmask, uint8_t data)
{
	uint8_t high = bitmask | flip[data>>4];
	uint8_t low = bitmask | flip[data & 0x0F];
	
	uint8_t return_data[4] = { high, 0b00100000, low | 0b00100000, low };
	
	return return_data;
}
uint8_t Write_Lcd(uint8_t data, uint8_t char_mode)
{
	if (ddrb & 0b00010000)
	{
		uint8_t low = (portb & 0b00000001) | 0b01000000;
		uint8_t high = low | 0b00100000;
		
		if (!TWIM_Start(0x20, TWIM_WRITE))
		{
			TWIM_Stop();
			return FALSE;
		}
		
		TWIM_Write(MCP2307_GPIOB);
		TWIM_Write(low);
		TWIM_Stop();
		
		while (1)
		{
			if (!TWIM_Start(0x20, TWIM_WRITE))
			{
				TWIM_Stop();
				return FALSE;
			}
			
			TWIM_Write(high);
			TWIM_Stop();
			
			if (!TWIM_Start(0x20, TWIM_READ))
			{
				TWIM_Stop();
				return FALSE;
			}
			
			uint8_t bits = TWIM_ReadNack();
			TWIM_Stop();
			
			if (!TWIM_Start(0x20, TWIM_WRITE))
			{
				TWIM_Stop();
				return FALSE;
			}
			
			TWIM_Write(MCP2307_GPIOB);
			TWIM_Write(low);
			TWIM_Write(high);
			TWIM_Write(low);
			TWIM_Stop();
			
			if ((bits & 0b00000010) == 0)
			{
				break;
			}
		}
		
		portb = low;
		ddrb &= 0b11101111;
		
		if (!TWIM_Start(0x20, TWIM_WRITE))
		{
			TWIM_Stop();
			return FALSE;
		}
		
		TWIM_Write(MCP2307_IODIRB);
		TWIM_Write(ddrb);
		TWIM_Stop();
	}
	
	uint8_t bitmask = portb & 0b00000001;
	if (char_mode)
	{
		bitmask |= 0b10000000;
	}
	
	// Single Byte
	uint8_t* output;
	
	if (char_mode)
	{
		output = Out_Four(bitmask, 97);
	}
	else
	{
		output = Out_Four(bitmask, data);	
	}
	
	if (!TWIM_Start(0x20, TWIM_WRITE))
	{
		TWIM_Stop();
		return FALSE;
	}
		
	TWIM_Write(MCP2307_GPIOB);
	TWIM_Write(output[0]);
	TWIM_Write(output[1]);
	TWIM_Write(output[2]);
	TWIM_Write(output[3]);
	TWIM_Stop();
	
	portb = output[3];
	
	if (!char_mode && (data = LCD_CLEARDISPLAY || data == LCD_RETURNHOME))
	{
		ddrb |= 0b00010000;
			
		if (!TWIM_Start(0x20, TWIM_WRITE))
		{
			TWIM_Stop();
			return FALSE;
		}
	
		TWIM_Write(MCP2307_IODIRB);
		TWIM_Write(ddrb);
		TWIM_Stop();
	}
	
	return TRUE;
}

uint8_t Init_Lcd()
{
		if (!TWIM_Init((uint32_t)100000))
		{
			return FALSE;
		}
		
		if (!TWIM_Start(0x20, TWIM_WRITE))
		{
			TWIM_Stop();
			return FALSE;
		}
		
		TWIM_Write(MCP23017_IOCON_BANK1);
		TWIM_Write(0);
		TWIM_Stop();
		
		if (!TWIM_Start(0x20, TWIM_WRITE))
		{
			TWIM_Stop();
			return FALSE;
		}

		TWIM_Write(0b00111111);
		TWIM_Write(ddrb);
		TWIM_Write(0b00111111);
		TWIM_Write(0b00000000);
		TWIM_Write(0b00000000);
		TWIM_Write(0b00000000);
		TWIM_Write(0b00000000);
		TWIM_Write(0b00000000);
		TWIM_Write(0b00000000);
		TWIM_Write(0b00000000);
		TWIM_Write(0b00000000);
		TWIM_Write(0b00000000);
		TWIM_Write(0b00111111);
		TWIM_Write(0b00000000);
		TWIM_Write(0b00000000);
		TWIM_Write(0b00000000);
		TWIM_Write(0b00000000);
		TWIM_Write(0b00000000);
		TWIM_Write(porta);
		TWIM_Write(portb);
		TWIM_Write(porta);
		TWIM_Write(portb);
		TWIM_Stop();
		
		if (!TWIM_Start(0x20, TWIM_WRITE))
		{
			TWIM_Stop();
			return FALSE;
		}
		
		TWIM_Write(MCP23017_IOCON_BANK0);
		TWIM_Write(0b10100000);
		TWIM_Stop();
		
		if (!Write_Lcd(0x33, FALSE))
		{
			return FALSE;
		}
		
		if (!Write_Lcd(0x32, FALSE))
		{
			return FALSE;
		}
		
		if (!Write_Lcd(0x28, FALSE))
		{
			return FALSE;
		}
		
		if (!Write_Lcd(LCD_CLEARDISPLAY, FALSE))
		{
			return FALSE;
		}
		
		if (!Write_Lcd(LCD_CURSORSHIFT | displayshift, FALSE))
		{
			return FALSE;
		}
		
		if (!Write_Lcd(LCD_ENTRYMODESET | displaymode, FALSE))
		{
			return FALSE;
		}
		
		if (!Write_Lcd(LCD_DISPLAYCONTROL | displaycontrol, FALSE))
		{
			return FALSE;
		}
		
		if (!Write_Lcd(LCD_RETURNHOME, FALSE))
		{
			return FALSE;
		}
		
		return TRUE;
}

uint8_t Stop_Lcd()
{
	porta = 0b11000000;
	portb = 0b00000001;
	
	if (!TWIM_Start(0x20, TWIM_WRITE))
	{
		TWIM_Stop();
		return FALSE;
	}
	
	TWIM_Write(MCP23017_IOCON_BANK1);
	TWIM_Write(0);
	TWIM_Stop();
	
	if (!TWIM_Start(0x20, TWIM_WRITE))
	{
		TWIM_Stop();
		return FALSE;
	}

	TWIM_Write(0b00111111);
	TWIM_Write(ddrb);
	TWIM_Write(0b00111111);
	TWIM_Write(0b00000000);
	TWIM_Write(0b00000000);
	TWIM_Write(0b00000000);
	TWIM_Write(0b00000000);
	TWIM_Write(0b00000000);
	TWIM_Write(0b00000000);
	TWIM_Write(0b00000000);
	TWIM_Write(0b00000000);
	TWIM_Write(0b00000000);
	TWIM_Write(0b00111111);
	TWIM_Write(0b00000000);
	TWIM_Write(0b00000000);
	TWIM_Write(0b00000000);
	TWIM_Write(0b00000000);
	TWIM_Write(0b00000000);
	TWIM_Write(porta);
	TWIM_Write(portb);
	TWIM_Write(porta);
	TWIM_Write(portb);
	TWIM_Stop();
	
	if (!TWIM_Start(0x20, TWIM_WRITE))
	{
		TWIM_Stop();
		return FALSE;
	}
	
	TWIM_Write(MCP23017_IOCON_BANK0);
	TWIM_Write(0b10100000);
	TWIM_Stop();
	
	if (!Write_Lcd(0x33, FALSE))
	{
		return FALSE;
	}
	
	if (!Write_Lcd(0x32, FALSE))
	{
		return FALSE;
	}
	
	if (!Write_Lcd(0x28, FALSE))
	{
		return FALSE;
	}
	
	if (!Write_Lcd(LCD_CLEARDISPLAY, FALSE))
	{
		return FALSE;
	}
	
	if (Write_Lcd(LCD_CURSORSHIFT | displayshift, FALSE))
	{
		return FALSE;
	}
	
	if (!Write_Lcd(LCD_ENTRYMODESET | displaymode, FALSE))
	{
		return FALSE;
	}
	
	if (!Write_Lcd(LCD_DISPLAYCONTROL | displaycontrol, FALSE))
	{
		return FALSE;
	}
	
	if (!Write_Lcd(LCD_RETURNHOME, FALSE))
	{
		return FALSE;
	}
	
	return TRUE;
}

uint8_t Clear_Lcd()
{
	return Write_Lcd(LCD_CLEARDISPLAY, FALSE);
}

void Begin_Lcd(uint8_t cols, uint8_t lines)
{
	numlines = lines;
	Clear_Lcd();
}

//void Message_Lcd(char* text)
//{
	//char* token;
	//char* search = " ";
	//
	//token = strtok(text, search);
	//
//}

uint8_t Char_Lcd(char character)
{
	return Write_Lcd(character, TRUE);
}

uint8_t Backlight_Lcd(uint8_t color)
{
	uint8_t c = ~color;
	porta = (porta & 0b00111111) | ((c & 0b011) << 6);
	portb = (portb & 0b11111110) | ((c & 0b100) >> 2);
	
	if (!TWIM_Start(0x20, TWIM_WRITE))
	{
		TWIM_Stop();
		return FALSE;
	}
		
	TWIM_Write(MCP2307_GPIOA);
	TWIM_Write(porta);
	TWIM_Stop();
	
	if (!TWIM_Start(0x20, TWIM_WRITE))
	{
		TWIM_Stop();
		return FALSE;
	}
		
	TWIM_Write(MCP2307_GPIOB);
	TWIM_Write(portb);
	TWIM_Stop();
	
	return TRUE;
}

uint8_t Blink_Lcd()
{
	displaycontrol |= LCD_BLINKON;
	return Write_Lcd(LCD_DISPLAYCONTROL | displaycontrol, FALSE);
}

uint8_t Blink_Off_Lcd()
{
	displaycontrol &= ~LCD_BLINKON;
	return Write_Lcd(LCD_DISPLAYCONTROL | displaycontrol, FALSE);
}